import { ISession, IAPIClient, Tab } from "chrome-debugging-client";
import { Page } from "./debugging-protocol-domains";
import { Trace } from "./trace";
export interface BenchmarkMeta {
    browserVersion: string;
    cpus: string[];
}
export interface IBenchmark<State, Result> {
    name: string;
    /** convenience to run the benchmark, returning result */
    run(iterations: number): Promise<Result>;
    /** alternatively, the following can be invoked manually in order */
    /** setup the benchmark */
    setup(session: ISession): Promise<State>;
    /** run a single iteration of the benchmark */
    perform(session: ISession, state: State, iteration: number): Promise<State>;
    /** finalize the benchmark, returning result */
    finalize(session: ISession, state: State): Promise<Result>;
}
export interface ITab {
    /** The process id of the tab */
    pid: number;
    /** The current frame for the tab */
    frame: Page.Frame;
    /** Add a script to execute on load */
    addScriptToEvaluateOnLoad(source: string): Promise<Page.ScriptIdentifier>;
    /** Remove a previously added script */
    removeScriptToEvaluateOnLoad(identifier: Page.ScriptIdentifier): Promise<void>;
    /** Navigates to the specified url */
    navigate(url: string, waitForLoad?: boolean): Promise<void>;
    /** Start tracing */
    startTracing(categories: string, options?: string): Promise<ITracing>;
    /** Clear browser cache and memory cache */
    clearBrowserCache(): Promise<void>;
    /** Perform GC */
    collectGarbage(): Promise<void>;
}
export interface ITracing {
    traceComplete: Promise<Trace>;
    /** end early, normally records for duration or until buffer full  */
    end(): Promise<void>;
}
export interface BrowserOptions {
    type: string;
    executablePath?: string;
    chromiumSrcDir?: string;
}
export declare class Runner<R> {
    private benchmarks;
    constructor(benchmarks: IBenchmark<any, R>[]);
    run(iterations: number): Promise<R[]>;
    private inSequence<T>(callback);
}
export interface BenchmarkParams {
    name: string;
    browser: BrowserOptions;
}
export interface BenchmarkState<R> {
    apiClient: IAPIClient;
    tab: Tab;
    results: R;
}
export declare abstract class Benchmark<R> implements IBenchmark<BenchmarkState<R>, R> {
    name: string;
    private browserOptions;
    constructor(params: BenchmarkParams);
    protected abstract createResults(meta: BenchmarkMeta): R;
    protected warm(t: ITab): Promise<void>;
    protected abstract performIteration(t: ITab, results: R, index: number): Promise<void>;
    run(iterations: number): Promise<R>;
    setup(session: ISession): Promise<BenchmarkState<R>>;
    perform(session: ISession, state: BenchmarkState<R>, iteration: number): Promise<BenchmarkState<R>>;
    finalize(session: ISession, state: BenchmarkState<R>): Promise<R>;
    private withTab<T>(session, state, callback);
}
