"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator.throw(value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const chrome_debugging_client_1 = require("chrome-debugging-client");
const debugging_protocol_domains_1 = require("./debugging-protocol-domains");
const trace_1 = require("./trace");
const os = require("os");
function delay(ms) {
    return new Promise(resolve => setTimeout(resolve, ms));
}
function createSessions(count, callback) {
    if (count === 1) {
        return chrome_debugging_client_1.createSession(session => callback([session]));
    }
    else {
        return createSessions(count - 1, (sessions) => chrome_debugging_client_1.createSession(session => callback([session].concat(sessions))));
    }
}
class Runner {
    constructor(benchmarks) {
        this.benchmarks = benchmarks;
    }
    run(iterations) {
        return __awaiter(this, void 0, void 0, function* () {
            let benchmarks = this.benchmarks;
            return createSessions(benchmarks.length, (sessions) => __awaiter(this, void 0, void 0, function* () {
                let states = yield this.inSequence((benchmark, i) => benchmark.setup(sessions[i]));
                for (let iteration = 0; iteration < iterations; iteration++) {
                    states = yield this.inSequence((benchmark, i) => benchmark.perform(sessions[i], states[i], iteration));
                }
                return this.inSequence((benchmark, i) => benchmark.finalize(sessions[i], states[i]));
            }));
        });
    }
    inSequence(callback) {
        return __awaiter(this, void 0, void 0, function* () {
            let benchmarks = this.benchmarks;
            let results = [];
            for (let i = 0; i < benchmarks.length; i++) {
                results.push(yield callback(benchmarks[i], i));
            }
            return results;
        });
    }
}
exports.Runner = Runner;
class Benchmark {
    constructor(params) {
        this.name = params.name;
        this.browserOptions = params.browser;
    }
    warm(t) {
        return __awaiter(this, void 0, void 0, function* () { });
    }
    ;
    run(iterations) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Runner([this]).run(iterations)[0];
        });
    }
    // create session, spawn browser, get port connect to API to get version
    setup(session) {
        return __awaiter(this, void 0, void 0, function* () {
            let browserOptions = this.browserOptions;
            let browser = yield session.spawnBrowser(browserOptions.type, browserOptions);
            let apiClient = yield session.createAPIClient("localhost", browser.remoteDebuggingPort);
            let version = yield apiClient.version();
            let tabs = yield apiClient.listTabs();
            // open a blank tab
            let tab = yield apiClient.newTab("about:blank");
            for (let i = 0; i < tabs.length; i++) {
                yield apiClient.closeTab(tabs[i].id);
            }
            yield delay(1500);
            let browserVersion = version["Browser"];
            let cpus = os.cpus().map((cpu) => cpu.model);
            let results = this.createResults({ browserVersion, cpus });
            let state = { apiClient, tab, results };
            yield this.withTab(session, state, tab => this.warm(tab));
            return state;
        });
    }
    perform(session, state, iteration) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.withTab(session, state, tab => this.performIteration(tab, state.results, iteration));
            return state;
        });
    }
    finalize(session, state) {
        return __awaiter(this, void 0, void 0, function* () {
            return state.results;
        });
    }
    withTab(session, state, callback) {
        return __awaiter(this, void 0, void 0, function* () {
            let apiClient = state.apiClient;
            let tab = state.tab;
            let client = yield session.openDebuggingProtocol(tab.webSocketDebuggerUrl);
            let page = new debugging_protocol_domains_1.Page(client);
            yield page.enable();
            let res = yield page.getResourceTree();
            let frame = res.frameTree.frame;
            let dsl = new TabDSL(tab.id, client, page, frame);
            yield dsl.clearBrowserCache();
            yield dsl.collectGarbage();
            yield apiClient.activateTab(tab.id);
            yield delay(500);
            yield apiClient.activateTab(tab.id);
            let rtn = yield callback(dsl);
            state.tab = yield apiClient.newTab("about:blank");
            yield apiClient.closeTab(tab.id);
            return rtn;
        });
    }
}
exports.Benchmark = Benchmark;
class TabDSL {
    constructor(id, client, page, frame) {
        this.client = client;
        this.page = page;
        this.frame = frame;
        this.tracing = new debugging_protocol_domains_1.Tracing(client);
        this.network = new debugging_protocol_domains_1.Network(client);
        this.heapProfiler = new debugging_protocol_domains_1.HeapProfiler(client);
        this.currentTrace = null;
        let pid = this.pid = parseInt(frame.id.split(".")[0], 10);
        page.frameNavigated = (params) => {
            let frame = params.frame;
            if (this.frame.id = frame.id) {
                this.onNavigated(frame);
            }
        };
    }
    onNavigated(frame) {
        if (!frame.parentId) {
            this.frame = frame;
            // navigating to "about:blank" a signal to end the current trace
            if (this.currentTrace && frame.url === "about:blank") {
                this.currentTrace.end();
            }
        }
    }
    /** Navigates to the specified url */
    navigate(url, waitForLoad) {
        return __awaiter(this, void 0, void 0, function* () {
            let didLoad;
            if (waitForLoad) {
                didLoad = new Promise(resolve => {
                    this.page.frameStoppedLoading = params => {
                        if (params.frameId === this.frame.id) {
                            this.page.frameStoppedLoading = null;
                            resolve();
                        }
                    };
                });
            }
            if (this.frame.url === url) {
                yield this.page.reload({});
            }
            else {
                yield this.page.navigate({ url });
            }
            if (waitForLoad) {
                yield didLoad;
            }
        });
    }
    addScriptToEvaluateOnLoad(source) {
        return __awaiter(this, void 0, void 0, function* () {
            let result = yield this.page.addScriptToEvaluateOnLoad({ scriptSource: source });
            return result.identifier;
        });
    }
    removeScriptToEvaluateOnLoad(identifier) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.page.removeScriptToEvaluateOnLoad({ identifier });
        });
    }
    /** Start tracing */
    startTracing(categories, options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.currentTrace) {
                throw new Error("already tracing");
            }
            let tracing = this.tracing;
            yield tracing.start({
                categories: categories
            });
            let trace = this.currentTrace = new TracingDSL(tracing, this.pid);
            trace.traceComplete.then(() => {
                this.currentTrace = null;
            });
            return trace;
        });
    }
    /** Clear browser cache and memory cache */
    clearBrowserCache() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.network.enable();
            let res = yield this.network.canClearBrowserCache();
            if (!res.result) {
                throw new Error("Cannot clear browser cache");
            }
            yield this.network.clearBrowserCache();
            // causes MemoryCache entries to be evicted
            yield this.network.setCacheDisabled({ cacheDisabled: true });
            yield this.network.disable();
        });
    }
    collectGarbage() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.heapProfiler.enable();
            yield this.heapProfiler.collectGarbage();
            yield this.heapProfiler.disable();
        });
    }
}
class TracingDSL {
    constructor(tracing, pid) {
        this.tracing = tracing;
        this.isTracing = true;
        this.isComplete = false;
        let trace = new trace_1.Trace();
        tracing.dataCollected = (evt) => {
            trace.addEvents(evt.value);
        };
        this.traceComplete = new Promise((resolve) => {
            // TODO setTimeout
            tracing.tracingComplete = () => {
                tracing.dataCollected = null;
                tracing.tracingComplete = null;
                this.isTracing = false;
                this.isComplete = true;
                trace.mainProcess = trace.processMap[pid];
                resolve(trace);
            };
        });
    }
    end() {
        if (!this.endPromise) {
            if (this.isTracing) {
                this.endPromise = this.tracing.end().then(() => {
                    this.isTracing = false;
                });
            }
            else {
                this.endPromise = Promise.resolve();
            }
        }
        return this.endPromise;
    }
}
//# sourceMappingURL=benchmark.js.map