"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator.throw(value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const benchmark_1 = require("../benchmark");
const trace_event_1 = require("../trace/trace_event");
class InitialRenderMetric {
    constructor(events, markers) {
        this.events = events;
        this.markers = markers;
        this.start = 0;
        this.markerIdx = 0;
        this.lastMarkEvent = undefined;
        this.requests = new Map();
        this.seenAllMarks = false;
        this.done = false;
        this.stack = [];
        this.sample = {
            duration: 0,
            phaseSamples: [],
            js: 0,
            compile: 0,
            run: 0,
            gc: 0,
            gcSamples: [],
            net: 0,
            response: 0,
            requestSamples: []
        };
    }
    measure() {
        let events = this.events;
        for (let i = 0; i < events.length; i++) {
            if (this.done)
                break;
            let event = events[i];
            switch (event.ph) {
                case trace_event_1.TRACE_EVENT_PHASE_MARK:
                    if (!this.seenAllMarks) {
                        this.measureMark(event);
                    }
                    break;
                case trace_event_1.TRACE_EVENT_PHASE_COMPLETE:
                    this.measureComplete(event);
                    break;
                case trace_event_1.TRACE_EVENT_PHASE_BEGIN:
                    this.measureBegin(event);
                    break;
                case trace_event_1.TRACE_EVENT_PHASE_END:
                    this.measureEnd(event);
                    break;
                case trace_event_1.TRACE_EVENT_PHASE_INSTANT:
                    this.measureInstant(event);
                    break;
            }
        }
        return this.sample;
    }
    measureInstant(event) {
        switch (event.name) {
            case "ResourceSendRequest":
                this.measureResourceSend(event);
                break;
            case "ResourceReceiveResponse":
                this.measureResourceRespond(event);
                break;
            case "ResourceFinish":
                this.measureResourceFinish(event);
                break;
        }
    }
    measureResourceSend(event) {
        let data = event.args["data"];
        let requestId = data && data["requestId"];
        if (!requestId)
            return;
        let request = {
            url: data && data["url"],
            send: event.ts,
            response: 0,
            finish: 0
        };
        this.requests.set(requestId, request);
    }
    measureResourceRespond(event) {
        let data = event.args["data"];
        let requestId = data && data["requestId"];
        let request = requestId && this.requests.get(requestId);
        if (!request)
            return;
        request.response = event.ts;
    }
    measureResourceFinish(event) {
        let data = event.args["data"];
        let requestId = data && data["requestId"];
        let request = requestId && this.requests.get(requestId);
        if (!request)
            return;
        request.finish = event.ts;
        let sample = {
            url: request.url,
            response: request.response - request.send,
            duration: request.finish - request.send
        };
        this.sample.net += sample.duration;
        this.sample.response += sample.response;
        this.sample.requestSamples.push(sample);
    }
    measureBegin(begin) {
        this.stack.push(begin);
    }
    measureEnd(end) {
        let begin = this.stack.pop();
        let args = {};
        Object.keys(begin.args).forEach(key => args[key] = begin.args[key]);
        Object.keys(end.args).forEach(key => args[key] = end.args[key]);
        this.measureComplete({
            "pid": begin.pid,
            "tid": begin.tid,
            "ts": begin.ts,
            "ph": "X",
            "cat": begin.cat,
            "name": begin.name,
            "args": args,
            "dur": end.ts - begin.ts,
            "tdur": end.tts - begin.tts,
            "tts": begin.tts
        });
    }
    measureComplete(event) {
        switch (event.name) {
            // js entry points from browser
            // it is parsing and running a script
            // or it is calling a function on an event
            case "v8.compile":
                this.measureCompile(event);
                break;
            case "v8.run":
            case "v8.callFunction":
                this.measureRun(event);
                break;
            case "MajorGC":
            case "MinorGC":
                this.measureGC(event);
                break;
            case "Paint":
                this.measurePaint(event);
                break;
        }
    }
    measureGC(event) {
        this.sample.gc += event.dur;
        this.sample.gcSamples.push({
            duration: event.dur,
            usedHeapSizeBefore: event.args["usedHeapSizeBefore"],
            usedHeapSizeAfter: event.args["usedHeapSizeAfter"]
        });
    }
    measureCompile(event) {
        this.sample.js += event.dur;
        this.sample.compile += event.dur;
    }
    measureRun(event) {
        this.sample.js += event.dur;
        this.sample.run += event.dur;
    }
    measurePaint(event) {
        if (this.seenAllMarks) {
            let marker = this.markers[this.markerIdx - 1];
            this.sample.phaseSamples.push({
                phase: marker.label,
                self: event.ts - this.lastMarkEvent.ts,
                cumulative: event.ts - this.start
            }, {
                phase: "paint",
                self: event.dur,
                cumulative: event.ts - this.start + event.dur
            });
            this.sample.duration = event.ts - this.start + event.dur;
            this.done = true;
        }
    }
    measureMark(event) {
        let marker = this.markers[this.markerIdx];
        if (marker.start !== event.name) {
            return;
        }
        if (this.lastMarkEvent) {
            let lastMarker = this.markers[this.markerIdx - 1];
            let sample = event.ts - this.lastMarkEvent.ts;
            this.sample.phaseSamples.push({
                phase: lastMarker.label,
                self: event.ts - this.lastMarkEvent.ts,
                cumulative: event.ts - this.start
            });
        }
        else {
            this.start = event.ts;
        }
        this.lastMarkEvent = event;
        if (++this.markerIdx === this.markers.length) {
            this.seenAllMarks = true;
        }
    }
}
class InitialRenderBenchmark extends benchmark_1.Benchmark {
    constructor(params) {
        validateParams(params);
        super(params);
        this.params = params;
    }
    createResults(meta) {
        return {
            meta: meta,
            set: this.name,
            samples: []
        };
    }
    performIteration(t, results, i) {
        return __awaiter(this, void 0, void 0, function* () {
            let url = this.params.url;
            let markers = this.params.markers;
            let tracing = yield t.startTracing("blink.user_timing,devtools.timeline,v8");
            yield t.navigate(url);
            let trace = yield tracing.traceComplete;
            if (!trace.mainProcess || !trace.mainProcess.mainThread) {
                return;
            }
            let mainThread = trace.mainProcess.mainThread;
            mainThread.events.sort((a, b) => a.ts - b.ts);
            // if (i === 0) {
            //   fs.writeFileSync("trace.json", JSON.stringify(mainThread.events, null, 2));
            // }
            let metric = new InitialRenderMetric(mainThread.events, markers);
            let sample = metric.measure();
            // log progress to stderr
            // TODO make some events or logger
            console.error(this.name, sample.duration, "µs");
            results.samples.push(sample);
        });
    }
}
exports.InitialRenderBenchmark = InitialRenderBenchmark;
function validateParams(params) {
    if (!params.markers || params.markers.length === 0) {
        params.markers = [{
                start: "fetchStart",
                label: "render"
            }];
    }
    if (!params.url) {
        throw new Error("url is required");
    }
}
//# sourceMappingURL=initial-render.js.map